﻿/* 
 * Capture Test NyARToolkitCSサンプルプログラム
 * --------------------------------------------------------------------------------
 * The MIT License
 * Copyright (c) 2008 nyatla
 * airmail(at)ebony.plala.or.jp
 * http://nyatla.jp/nyartoolkit/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * 
 */
using System;
using System.IO;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using NyARToolkitCSUtils.Capture;
using NyARToolkitCSUtils.Direct3d;
using NyARToolkitCSUtils.NyAR;
using jp.nyatla.nyartoolkit.cs;
using jp.nyatla.nyartoolkit.cs.core;
using jp.nyatla.nyartoolkit.cs.detector;

namespace PMDMikuDirect3d
{
    public partial class SimpleLiteD3d : IDisposable, CaptureListener
    {
        private const int SCREEN_WIDTH=640;
        private const int SCREEN_HEIGHT=480;
        private const String AR_CODE_FILE = "../../AR_Data/patt.hiro";
        private const String AR_CAMERA_FILE = "../../AR_Data/camera_para.dat";
        private CaptureDevice _cap;          //DirectShowからのキャプチャ
        private NyARSingleDetectMarker _ar;  //NyARToolkit
        private DsBGRX32Raster _raster;      //NyARToolkit
        private NyARD3dUtil _utils;          //NyARToolkit
        private NyARSurface_XRGB32 _surface; //背景テクスチャ
        private Device _device = null;       // Direct3D デバイス
        private NyARTransMatResult __OnBuffer_nyar_transmat = new NyARTransMatResult();
        private bool _is_marker_enable;      //NyARToolkit
        private Matrix _trans_mat;           //NyARToolkit

        private VertexBuffer _vertexBuffer = null; //【GPU側VertexBuffer】
        private IndexBuffer _indexBuffer = null;   //【GPU側IndexBuffer】
        private float rot3 = 0.0f;                 //【モデルを回転表示するために使用】           
        private int SkinNumber = 1; //【表示する表情の番号】0はbaseデータなので表情としては表示できない
        private float SkinEffect;

        //*******************************************************************************************
        //【非同期イベントハンドラ】 CaptureDeviceからのイベントをハンドリングして、バッファとテクスチャを更新する。
        //*******************************************************************************************
        public void OnBuffer(CaptureDevice i_sender, double i_sample_time, IntPtr i_buffer, int i_buffer_len)
        {
            int w = i_sender.video_width;
            int h = i_sender.video_height;
            int s = w * (i_sender.video_bit_count / 8);
            NyARTransMatResult nyar_transmat = this.__OnBuffer_nyar_transmat;
            
            //テクスチャにRGBを取り込み()
            lock (this)
            {
                //カメラ映像をARのバッファにコピー
                this._raster.setBuffer(i_buffer, i_sender.video_vertical_flip);
                //マーカーは見つかったかな？
                bool is_marker_enable = this._ar.detectMarkerLite(this._raster, 120);
                if (is_marker_enable)
                {
                    //あればMatrixを計算
                    this._ar.getTransmationMatrix(nyar_transmat);
                    this._utils.toD3dMatrix(nyar_transmat, ref this._trans_mat);
                }
                this._is_marker_enable=is_marker_enable;
                this._surface.CopyFromXRGB32(this._raster); //テクスチャ内容を更新
            }
            return;
        }

        //【キャプチャを開始する関数】
        public void StartCap()  {  this._cap.StartCapture(); return;  }
        //【キャプチャを停止する関数】
        public void StopCap()   {  this._cap.StopCapture();  return;  }

        //*******************************************************************************************
        //【Direct3Dデバイスを準備する関数】
        //*******************************************************************************************
        private Device PrepareD3dDevice(Control i_window)
        {
            PresentParameters pp = new PresentParameters();
            pp.Windowed = true;
            pp.SwapEffect = SwapEffect.Flip;
            pp.BackBufferFormat = Format.X8R8G8B8;
            pp.BackBufferCount = 1;
            pp.EnableAutoDepthStencil = true;
            pp.AutoDepthStencilFormat = DepthFormat.D16;
            CreateFlags fl_base = CreateFlags.FpuPreserve;

            try{
                return new Device(0, DeviceType.Hardware, i_window.Handle, fl_base|CreateFlags.HardwareVertexProcessing, pp);
            }catch (Exception ex1){
                Debug.WriteLine(ex1.ToString());
                try{
                    return new Device(0, DeviceType.Hardware, i_window.Handle, fl_base | CreateFlags.SoftwareVertexProcessing, pp);
                }catch (Exception ex2){
                    // 作成に失敗
                    Debug.WriteLine(ex2.ToString());
                    try{
                        return new Device(0, DeviceType.Reference, i_window.Handle, fl_base | CreateFlags.SoftwareVertexProcessing, pp);
                    }catch (Exception ex3){
                        throw ex3;
                    }
                }
            }
        }

        //*******************************************************************************************
        //【アプリケーションの初期化を行う】
        //*******************************************************************************************
        public bool InitializeApplication(Form1 topLevelForm, CaptureDevice i_cap_device)
        {
            topLevelForm.ClientSize=new Size(SCREEN_WIDTH,SCREEN_HEIGHT);
            //キャプチャを作る(QVGAでフレームレートは30)
            i_cap_device.SetCaptureListener(this);
            i_cap_device.PrepareCapture(SCREEN_WIDTH, SCREEN_HEIGHT, 15);
            this._cap = i_cap_device;
            
            //ARの設定

            //ARラスタを作る(DirectShowキャプチャ仕様)。
            this._raster = new DsBGRX32Raster(i_cap_device.video_width, i_cap_device.video_height, i_cap_device.video_width * i_cap_device.video_bit_count / 8);

            //AR用カメラパラメタファイルをロードして設定
            NyARParam ap = new NyARParam();
            ap.loadARParamFromFile(AR_CAMERA_FILE);
            ap.changeScreenSize(SCREEN_WIDTH, SCREEN_HEIGHT);

            //AR用のパターンコードを読み出し	
            NyARCode code = new NyARCode(16, 16);
            code.loadARPattFromFile(AR_CODE_FILE);

            //１パターンのみを追跡するクラスを作成
            this._ar = new NyARSingleDetectMarker(ap, code, 80.0, this._raster.getBufferType(), NyARSingleDetectMarker.PF_NYARTOOLKIT);
            
            //Direct3d用のユーティリティ準備
            this._utils = new NyARD3dUtil();

            //計算モードの設定
            this._ar.setContinueMode(true);

            //3dデバイスを準備する
            this._device = PrepareD3dDevice(topLevelForm);
            this._device.RenderState.ZBufferEnable = true;                // 陰面処理をｚバッファ方式で行う
            this._device.RenderState.ZBufferFunction = Compare.LessEqual; //.LessEqual;    
     //     this._device.RenderState.ZBufferFunction = Compare.Less; //【SphereMapping対応のMiku(metalic)では全身緑のメタリックになる】
            this._device.RenderState.Lighting = true;                     // 光源計算処理を行う　true;

            // 平行光源（無限遠に設定される方向を持った光源。太陽光の様な光源）
            this._device.Lights[0].Type = LightType.Directional;
            this._device.Lights[0].Ambient = Color.FromArgb(255, 200, 200, 200);  
            this._device.Lights[0].Diffuse = Color.White;
            this._device.Lights[0].Direction = new Vector3(-0.57735f, -0.57735f, -0.57735f);
            this._device.Lights[0].Enabled = true;
            // 環境光（全ての物体を均一に照らす光源）
            this._device.RenderState.Ambient = Color.FromArgb(0x202020);

            //カメラProjectionの設定
            Matrix tmp = new Matrix();
            this._utils.toCameraFrustumRH(ap, ref tmp);
            this._device.Transform.Projection = tmp;

            // ビュー変換の設定(左手座標系ビュー行列で設定する) (0,0,0)からZ+方向を向いてY軸が上方向
            this._device.Transform.View = Matrix.LookAtLH(
                new Vector3(0.0f, 0.0f, 0.0f), new Vector3(0.0f, 0.0f, 1.0f), new Vector3(0.0f, 1.0f, 0.0f));
            Viewport vp = new Viewport();
            vp.X = 0;
            vp.Y = 0;
            vp.Height = ap.getScreenSize().h;
            vp.Width = ap.getScreenSize().w;
            vp.MaxZ = 1.0f;
            //ビューポート設定
            this._device.Viewport = vp;

            //【頂点バッファの準備】頂点データ(X,Y,Z,Nx,Ny,Nz,Tu,Tv)を登録するバッファの準備
            this._vertexBuffer = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured),
                PMDv.nPoints, this._device, Usage.None, CustomVertex.PositionNormalTextured.Format, Pool.Managed);
            //【インデックスバッファの準備】第２引数はバッファの総バイト数＝(三角ポリゴンの数)*(ひとつの三角ポリゴンの頂点数)*(Int16は2bytes)
            this._indexBuffer = new IndexBuffer(this._device, PMDi.nPolygons * 3 * 2, Usage.WriteOnly, Pool.Managed, true);

            //【インデックスバッファをロックする】
            //●インデックスバッファは途中で変更することがないので初期化時にのみ転送
            //●頂点バッファは、表情/Bone処理で頂点座標を、Sphere MappingでTu,Tvを変更するのでMainLoop内で転送
            using (GraphicsStream data = this._indexBuffer.Lock(0, 0, LockFlags.None))
            {
                data.Write(PMDi.IndexBuffer); // インデックスデータをインデックスバッファにコピーします
                this._indexBuffer.Unlock();   // インデックスバッファのロックを解除します
            }
            //【テクスチャーの準備】PMDは材質ごとにTextureFileName[i]でテクスチャー使用の有無を判定
            PMDm.Texture = new Texture[PMDm.nMaterial];
            for (var i = 0; i < PMDm.nMaterial; i++)
            {
                if (PMDm.TextureFileName[i] != "")
                { PMDm.Texture[i] = TextureLoader.FromFile(this._device, PMDm.TextureFileName[i]); }
                else
                { PMDm.Texture[i] = null; }
            }

            //背景サーフェイスを作成
            this._surface = new NyARSurface_XRGB32(this._device, SCREEN_WIDTH, SCREEN_HEIGHT);

            this._is_marker_enable = false;
            return true;
        }

        //*******************************************************************************************
        //【メインループ処理】
        //*******************************************************************************************
        public void MainLoop()
        {
            //ARの計算
            lock (this)
            {
                // 背景サーフェイスを直接描画
                Surface dest_surface = this._device.GetBackBuffer(0, 0, BackBufferType.Mono);
                Rectangle src_dest_rect = new Rectangle(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
                this._device.StretchRectangle(this._surface.d3d_surface, src_dest_rect, dest_surface, src_dest_rect, TextureFilter.None);

                //【3Dオブジェクトの描画初期化】
                this._device.BeginScene(); //【レンダリングのシーン作成開始】
                this._device.Clear(ClearFlags.ZBuffer, Color.DarkBlue, 1.0f, 0); //【フレームバッファ・クリア】Zバッファも含む 
                //マーカーが見つかっていて、0.4より一致してたら描画する。
                if (this._is_marker_enable && this._ar.getConfidence()>0.4)
                {
                    this._device.SetStreamSource(0, this._vertexBuffer, 0); //【頂点バッファをデバイスのデータストリームにバインド】
                    this._device.VertexFormat = CustomVertex.PositionNormalTextured.Format; //【頂点バッファ書式をセット】
                    this._device.Indices = this._indexBuffer; //【インデックスバッファをセット】
                    this._device.RenderState.Lighting = true; //【光源計算処理を行う】
                    this._device.Lights[0].Enabled = true;    //【光源設定を有効にする】
                    this._device.RenderState.CullMode = Cull.None; //【袖やスカートが透明にならないようにカーリングを行わない】 
                    Matrix transform_mat2 = Matrix.Translation(0.0f, -250.0f, 0.0f); //【PMDのマーカーからの平行移動距離を設定】
                    //【PMDをx軸回りに回転表示させる処理】
                    Matrix transform_rot3 = Matrix.RotationYawPitchRoll( ( rot3 / 180.0f) * 3.141592f, 0.0f, 0.0f);
                    
                    rot3 += 2.0f; //【2°ずつ回転させる。1回転は180フレーム】
                    if (rot3 >= 360.0f)
                    {  rot3 = 0.0f;  
                       SkinNumber++; //【１周ごとに表情を変更】
                       if (SkinNumber >= PMDs.nSkin)
                       { SkinNumber = PMDs.nSkin; } //【baseは表情ではないので表情番号1から】 
                    }
                    SkinEffect = (float)((int)rot3 % 180)/180.0f; //【表情を表示してみる】
                    if (rot3 > 180.0f && rot3 <= 270.0f) { SkinEffect = 0.0f; }
                    if (rot3 > 270.0f && rot3 <= 360.0f) { SkinEffect = 1.0f; }

                    InitializeFacialExpression(); //【表情baseデータを表情ベクトルバッファSkinSetData.Positionに初期化】

                    if (SkinNumber < PMDs.nSkin)
                    {  transform_mat2 *= this._trans_mat; //【表情デモ用】【座標変換マトリックスを計算】
                       SetFacialExpression(SkinNumber, SkinEffect); //【すべての表情をデモ】
                    }
                    else
                    { //【にっこりと笑う】
                        transform_mat2 *= transform_rot3 * this._trans_mat; //【回転デモ用】【座標変換マトリックスを計算】
                        SkinEffect =  0.4f; SetFacialExpression("にこり", SkinEffect); //【表情】 3:(眉)
                        SkinEffect =  1.0f; SetFacialExpression("笑い", SkinEffect);   //【表情】 8:(眼)
                        SkinEffect = 0.35f; SetFacialExpression("あ", SkinEffect);     //【表情】19:(リップ)
                    }
                    TransferFacialExpression(); //【表情ベクトルをPMDv.UserVertexに転送】
                    this._device.SetTransform(TransformType.World, transform_mat2); //【上記マトリックスで座標変換を行う】

                    if (PMDm.SphereFlag)
                    { //【Sphere Mappingを使っているモデル】
                        int j = 0; int k; int sw;                           //【Traverse関連】[sw] 0:通常 1:ﾃｸｽﾁｬｰ 2:Sphere
                        Matrix matWorld, matView, matWV;                    //【SphereMap関連】
                        float m11, m12, m13, m21, m22, m23, m31, m32, m33;  //【SphereMap関連】
                        float nx, ny, nz;                                   //【SphereMap関連】
                        for (int i = 0; i < PMDm.nMaterial; i++)
                        {
                            this._device.Material = PMDm.Material[i];
                            k = PMDm.nPoints[i];
                            this._device.SetTexture(0, null);
                            sw = 0;
                            if (PMDm.TextureFileName[i] != "")
                            {
                                sw = 1; if (PMDm.SphereMap[i]) { sw = 2; }
                            }

                            switch (sw)
                            {
                                case 0: //【Polygon Fill】通常のポリゴン描画
                                    this._device.SetTexture(0, null);
                                    //【頂点バッファをロックする】
                                    //●インデックスバッファは途中で変更することがないので初期化時にのみ転送
                                    //●頂点バッファは、表情/Bone処理で頂点座標を、Sphere MappingでTu,Tvを変更するのでMainLoop内で転送
                                    using (GraphicsStream data = this._vertexBuffer.Lock(0, 0, LockFlags.None))
                                    {
                                        data.Write(PMDv.OriginalVertex); // 頂点データを頂点バッファにコピーします
                                        this._vertexBuffer.Unlock(); // 頂点バッファのロックを解除します
                                    }
                                    break;
                                case 1: //【Texture Mapping】
                                    this._device.SetTexture(0, PMDm.Texture[i]);
                                    //【頂点バッファをロックする】
                                    //●インデックスバッファは途中で変更することがないので初期化時にのみ転送
                                    //●頂点バッファは、表情/Bone処理で頂点座標を、Sphere MappingでTu,Tvを変更するのでMainLoop内で転送
                                    using (GraphicsStream data = this._vertexBuffer.Lock(0, 0, LockFlags.None))
                                    {
                                        data.Write(PMDv.OriginalVertex); // 頂点データを頂点バッファにコピーします
                                        this._vertexBuffer.Unlock(); // 頂点バッファのロックを解除します
                                    }
                                    break;
                                case 2: //【Sphere Mapping】
                                    this._device.SetTexture(0, PMDm.Texture[i]);
                                    // 現在のワールドビュー座標を取得
                                    matView = this._device.GetTransform(TransformType.View);
                                    matWorld = this._device.GetTransform(TransformType.World);
                                    matWV = matWorld * matView;
                                    // 処理速度を上げるためにワールドビュー座標の要素を抽出
                                    m11 = matWV.M11; m21 = matWV.M21; m31 = matWV.M31;
                                    m12 = matWV.M12; m22 = matWV.M22; m32 = matWV.M32;
                                    m13 = matWV.M13; m23 = matWV.M23; m33 = matWV.M33;
                                    // 各頂点に対してをループし、トランスフォームと正しいテクスチャ座標の計算を行う
                                    int jj;
                                    for (int ii = j; ii < (j + k); ii++)
                                    {
                                        jj = ii % PMDv.nPoints;  //【VertexBufferを何回もなめているようだ。MMDはAlphaBlendしてる？】
                                        nx = PMDv.UserVertex[jj].Normal.X;
                                        ny = PMDv.UserVertex[jj].Normal.Y;
                                        nz = PMDv.UserVertex[jj].Normal.Z;
                                        // zコンポーネントをチェックして後ろ側を向いている頂点をスキップ
                                        if (nx * m13 + ny * m23 + nz * m33 > 0.0f)
                                        {   // スフィアマップのテクスチャ座標を割り当てる
                                            PMDv.UserVertex[jj].Tu = 0.5f * (1.0f + (nx * m11 + ny * m21 + nz * m31));
                                            PMDv.UserVertex[jj].Tv = 0.5f * (1.0f - (nx * m12 + ny * m22 + nz * m32));
                                        }
                                    }
                                    //【頂点バッファをロックする】
                                    //●インデックスバッファは途中で変更することがないので初期化時にのみ転送
                                    //●頂点バッファは、表情/Bone処理で頂点座標を、Sphere MappingでTu,Tvを変更するのでMainLoop内で転送
                                    using (GraphicsStream data = this._vertexBuffer.Lock(0, 0, LockFlags.None))
                                    {
                                        data.Write(PMDv.UserVertex); // 頂点データを頂点バッファにコピーします
                                        this._vertexBuffer.Unlock(); // 頂点バッファのロックを解除します
                                    }
                                    break;
                            }
                            this._device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, PMDv.nPoints - 1, j, k);
                            j += k;
                        }

                    }
                    else //【通常のポリゴン・フィルとテクスチャーマッピングだけのモデル】
                    {
                        //【頂点バッファをロックする】
                        //●インデックスバッファは途中で変更することがないので初期化時にのみ転送
                        //●頂点バッファは、表情/Bone処理で頂点座標を、Sphere MappingでTu,Tvを変更するのでMainLoop内で転送
                        using (GraphicsStream data = this._vertexBuffer.Lock(0, 0, LockFlags.None))
                        {
                            data.Write(PMDv.UserVertex); // 頂点データを頂点バッファにコピーします
                            this._vertexBuffer.Unlock(); // 頂点バッファのロックを解除します
                        }
                        int j = 0; int k; //【Traverse関連】
                        for (int i = 0; i < PMDm.nMaterial; i++)
                        {
                            this._device.Material = PMDm.Material[i];
                            k = PMDm.nPoints[i];
                            if (PMDm.TextureFileName[i] != "") {  this._device.SetTexture(0, PMDm.Texture[i]);  }
                            else                               {  this._device.SetTexture(0, null);             }
                            this._device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, PMDv.nPoints - 1, j, k);
                            j += k;
                        }
                    }

                }



                this._device.EndScene(); //【レンダリングのシーンここまで】
                this._device.Present();  //【実際のディスプレイに描画】
            }
            return;
        }

        // リソースの破棄をするために呼ばれる
        public void Dispose()
        {
            lock (this)
            {
                // 頂点バッファを解放
                if (this._vertexBuffer != null)  {  this._vertexBuffer.Dispose();  }
                // インデックスバッファを解放
                if (this._indexBuffer != null)   {  this._indexBuffer.Dispose();  }
                if (this._surface != null)       {  this._surface.Dispose();      }
                // Direct3D デバイスのリソース解放
                if (this._device != null)        {  this._device.Dispose();       }
            }
        }
    }
}
